<?php
namespace App\Handlers;

use Illuminate\Support\Facades\URL;
use App\Helpers\Traits\BookXmlFilesTrait;
/**
 * TextMorphologyHandler.php
 *
 * @author: Leonard Smith <leonard@acornwebconsultants.com>
 * Date: 8/18/20
 * Time: 7:31 PM
 */

class TextMorphologyHandler
{
    use BookXmlFilesTrait;

    public function getVerse($book, $chapter, $verse)
    {
        $document = $this->openBookXml($book);

        $verseNodes = $document->getElementsByTagName('verse');

        $words = [];
        foreach ($verseNodes as $vn) {
            if ($vn->getAttribute('name') === ucfirst($book) . ' ' . $chapter . ':' . $verse) {
                $wordNodes = $vn->getElementsByTagName('w');

                foreach ($wordNodes as $w) {
                    $words[] = [
                        'ulb' => $w->nodeValue,
                        'greek' => $w->getAttribute('text'),
                        'morph' => $w->getAttribute('morph'),
                        'ognt-sort' => $w->getAttribute('OGNTsort'),
                        'lexeme' => $w->getAttribute('lexeme'),
                    ];
                }
            }
        }

        return $words;
    }

    public function getWordsJson($book, $chapter, $verse)
    {
        $words = $this->getVerse($book, $chapter, $verse);

        $json = [];
        foreach ($words as $word) {
            $json[] = $this->getWordJson(implode('-', [$book, $chapter, $verse, $word['ognt-sort']]), $word);
        }

        return json_encode($json);
    }

    protected function getWordJson($id, $word)
    {
        return [
            'type' => 'words',
            'id' => $id,
            'attributes' => $word,
            'links' => [
                'self' => URL::route('verse-word', $id),
            ]
        ];
    }

    public function getWordById($id)
    {
        $idParts = explode('-', $id);

        $book = $idParts[0];
        $chapter = $idParts[1];
        $verse = $idParts[2];
        $ogntSort = $idParts[3];

        $words = $this->getVerse($book, $chapter, $verse);

        $str = [];
        foreach ($words as $word) {
            $str[] = $word;
            if ($word['ognt-sort'] == $ogntSort) {
                return $this->getWordJson($id, $word);
            }
        }
        return "NOT FOUND: $ogntSort\n";
    }

    protected function openBookXml($book)
    {
        $book = strtolower($book);

        $xmlFile = $this->getBookXmlFile('/ulb', $book);

        $document = new \DOMDocument;
        $document->load($xmlFile);

        return $document;
    }

    public function getChaptersByBook($book)
    {
        $document = $this->openBookXml($book);

        $chapterNodes = $document->getElementsByTagName('chapter');

        $data = [];
        foreach ($chapterNodes as $node) {
            $data[] = $this->getChapterDataFromNode($node);
        }

        return json_encode($data);
    }

    public function getChapterDataById($id)
    {

    }

    public function getChapterDataFromNode($node)
    {
        return [
            'chapter' => $this->getChapterNameFromOsisId($node->getAttribute('osisID')),
            'verses' => $this->getUlbChapterVerses($node),
        ];
    }

    public function getChapterNameFromOsisId($osisId)
    {
        preg_match("|^[A-Za-z1-3]*.([0-9]*)$|", $osisId, $matches);

        return $matches[1];
    }

    public function getUlbChapterVerses($xmlNode)
    {
        $verseNodes = $xmlNode->getElementsByTagName('verse');

        $verses = [];
        foreach ($verseNodes as $vn) {
            $verse['number'] = $this->getVerseNumber($vn->getAttribute('name'));
            $verse['text'] = $vn->getElementsByTagName('ULB')->item(0)->nodeValue;

            $verses[] = $verse;
        }

        return $verses;
    }

    protected function getVerseNumber($string)
    {
        preg_match("|^[A-Za-z1-3]*\s*[0-9]*:([0-9]*)$|", $string, $matches);

        return $matches[1];
    }
}
